﻿// 32feet.NET - Personal Area Networking for .NET
//
// InTheHand.Net.Bluetooth.Widcomm.WidcommBluetoothFactoryBase
// 
// Copyright (c) 2010 Alan J McFarlane, All rights reserved.
// Copyright (c) 2010 In The Hand Ltd, All rights reserved.
// This source code is licensed under the In The Hand Community License - see License.txt

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO.Ports;
using InTheHand.Net.Bluetooth.Factory;
using InTheHand.Net.Sockets;


namespace InTheHand.Net.Bluetooth.BlueSoleil
{
    class BluesoleilClient : CommonDiscoveryBluetoothClient
    {
        static readonly int? BufferSize = null; // = 64 * 1024;
        //
        readonly BluesoleilFactory _factory;
        readonly Action<BluetoothEndPoint> _beginConnectDlgt;
        //
        UInt32? _hDev;
        SerialPortNetworkStream _stream;
        BluetoothEndPoint _remoteEp;

        //----
        internal BluesoleilClient(BluesoleilFactory fcty)
        {
            if (fcty == null) throw new ArgumentNullException("fcty");
            _factory = fcty;
            _beginConnectDlgt = new Action<BluetoothEndPoint>(Connect);
        }

        [SuppressMessage("Microsoft.Usage", "CA1801:ReviewUnusedParameters", MessageId = "localEP")]
        internal BluesoleilClient(BluesoleilFactory fcty, BluetoothEndPoint localEP)
            : this(fcty)
        {
            Debug.Fail("Ignoring localEP");
        }

        //----
        protected override void Dispose(bool disposing)
        {
            if (disposing) {
                if (_stream != null) _stream.Close();
            }
            // TODO Kill DeviceDiscovery etc
        }

        //----
        protected override void BeginInquiry(int maxDevices, AsyncCallback callback, object state,
            BluetoothClient.LiveDiscoveryCallback liveDiscoHandler, object liveDiscoState)
        {
            _factory.BeginInquiry(maxDevices, InquiryLength,
                callback, state,
                liveDiscoHandler, liveDiscoState);
        }

        protected override List<IBluetoothDeviceInfo> GetKnownRemoteDeviceEntries()
        {
            bool authenticated = true;
            bool remembered = true;
            List<IBluetoothDeviceInfo> knownDevices = _factory.GetRememberedDevices(
                authenticated, remembered);
            return knownDevices;
        }

        protected override List<IBluetoothDeviceInfo> EndInquiry(IAsyncResult ar)
        {
            List<IBluetoothDeviceInfo> devices = _factory.EndInquiry(ar);
            return devices;
        }

        //--------------------------------------------------------------
        public override System.Net.Sockets.Socket Client
        {
            get { throw new NotSupportedException("This stack does not use Sockets."); }
            set { throw new NotSupportedException("This stack does not use Sockets."); }
        }

        //----
        public override void Connect(BluetoothEndPoint remoteEP)
        {
            if (remoteEP.Port != 0 && remoteEP.Port != -1)
                throw new NotSupportedException("Don't support connect to particular port.  Please can someone tell me how.");
            //
            _factory.SdkInit();
            _factory.RegisterCallbacksOnce(); // Need to use event DISC_IND to close the Stream/SerialPort.
            //
            BluesoleilDeviceInfo bdi = BluesoleilDeviceInfo.CreateFromGivenAddress(remoteEP.Address, _factory);
            UInt32 hDev = bdi.Handle;
            UInt32 hConn;
            Structs.BtSdkAppExtSPPAttrStru sppAttr = new Structs.BtSdkAppExtSPPAttrStru(remoteEP);
            //
            Debug.WriteLine("Gonna Btsdk_ConnectAppExtSPPService with: " + sppAttr.ToString());
            BtSdkError ret = _factory.Api.Btsdk_ConnectAppExtSPPService(hDev, ref sppAttr, out hConn);
            Debug.WriteLine(string.Format(CultureInfo.InvariantCulture,
                "ret: {0}, hConn: 0x{1:X}, with: {2}", ret, hConn, sppAttr.ToString()));
            _hDev = hDev;
            BluesoleilUtils.CheckAndThrow(ret, "Btsdk_ConnectAppExtSPPService");
            Debug.Assert(sppAttr.rf_svr_chnl != 0, "sppAttr.rf_svr_chnl: " + sppAttr.rf_svr_chnl);
            if (sppAttr.rf_svr_chnl == 0) {
                // Highly likely an OPP/etc connection was made, and not a RFCOMM
                // connection, and thus no COM port we can use. :-(  So fail!
                Trace.WriteLine(string.Format(CultureInfo.InvariantCulture,
                    "BlueSoleil seems no RFCOMM connection made, closing. (r: {0}, sp: {1}",
                    sppAttr.rf_svr_chnl, sppAttr.com_index));
                Debug.Assert(sppAttr.com_index == 0, "Don't expect a COM port to be created in this (fail) case.");
                // (Note: Add a dummy record so RemoveLiveConnection works ok).
                int liveCountB = _factory.AddConnection(hConn, NullBluesoleilConnection.Instance);
                Debug.WriteLine(string.Format(CultureInfo.InvariantCulture,
                    "BlueSoleilClient.Connect non-RFCOMM LiveConns count: {0}.", liveCountB));
                var retD = _factory.Api.Btsdk_Disconnect(hConn);
                BluesoleilUtils.Assert(retD, "Close non-RFCOMM connection");
                throw BluesoleilUtils.ErrorConnectIsNonRfcomm();
            }
            _remoteEp = new BluetoothEndPoint(remoteEP.Address, BluetoothService.Empty, sppAttr.rf_svr_chnl);
            //
            int comPort = sppAttr.com_index;
            //System.Windows.Forms.MessageBox.Show("BlueSoleil created port: '" + comPort + "'.");
            var serialPort = new System.IO.Ports.SerialPort();
            if (BufferSize.HasValue) {
                serialPort.ReadBufferSize = BufferSize.Value;
                serialPort.WriteBufferSize = BufferSize.Value;
            }
            serialPort.PortName = "COM" + comPort;
            serialPort.Handshake = Handshake.RequestToSend;
            serialPort.Open();
            //((System.ComponentModel.IComponent)serialPort).Disposed
            //    += delegate { System.Windows.Forms.MessageBox.Show("BlueSoleil's SerialPort disposed"); };
            //
            // We pass '_hConn' as the Stream handles calling
            // Btsdk_Disconnect to close the RFCOMM connection.
            var strm = new BlueSoleilSerialPortNetworkStream(serialPort, hConn, this, _factory);
            _stream = strm;
            int liveCount = _factory.AddConnection(hConn, strm);
            Debug.WriteLine(string.Format(CultureInfo.InvariantCulture,
                "BlueSoleilClient.Connect LiveConns count: {0}.", liveCount));
            // TODO who handles closing the connection if opening the port fails
        }

        public override IAsyncResult BeginConnect(BluetoothEndPoint remoteEP, AsyncCallback requestCallback, object state)
        {
            return _beginConnectDlgt.BeginInvoke(remoteEP, requestCallback, state);
        }

        public override void EndConnect(IAsyncResult asyncResult)
        {
            _beginConnectDlgt.EndInvoke(asyncResult);
        }

        public override bool Connected
        {
            get
            {
                var strm = _stream;
                if (strm == null) return false;
                if (!_hDev.HasValue) {
                    // When does ths occur?????
                    //Don't need it likely.........return false;
                }
                return strm.Connected;
            }
        }

        public override int Available
        {
            get
            {
                SerialPortNetworkStream strm = _stream;
                if (strm == null) throw new InvalidOperationException("Not connected.");
                return strm.Available;
            }
        }

        public override bool Authenticate
        {
            get { return false; }
            set
            {
                throw new NotSupportedException("BlueSoleil does not support setting authentication/encryption.");
            }
        }

        public override bool Encrypt
        {
            get { return false; }
            set
            {
                throw new NotSupportedException("BlueSoleil does not support setting authentication/encryption.");
            }
        }

        public override string GetRemoteMachineName(BluetoothAddress device)
        {
            IBluetoothDeviceInfo bdi = _factory.DoGetBluetoothDeviceInfo(device);
            return bdi.DeviceName;
        }

        public override System.Net.Sockets.NetworkStream GetStream()
        {
            if (_stream == null)
                throw new InvalidOperationException("Not connected.");
            return _stream;
        }

        public override System.Net.Sockets.LingerOption LingerState
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        //----
        public override Guid LinkKey
        {
            get { throw new NotImplementedException(); }
        }

        public override LinkPolicy LinkPolicy
        {
            get { throw new NotImplementedException(); }
        }

        public override BluetoothEndPoint RemoteEndPoint
        {
            get
            {
                if (!Connected || _remoteEp == null)
                    throw new InvalidOperationException("Not connected.");
                return _remoteEp;
            }
        }

        public override string RemoteMachineName
        {
            get
            {
                if (!Connected || _hDev == null)
                    throw new InvalidOperationException("Not connected.");
                IBluetoothDeviceInfo bdi = BluesoleilDeviceInfo
                    .CreateFromHandleFromConnection(_hDev.Value, _factory);
                return bdi.DeviceName;
            }
        }

        //----
        public override void SetPin(string pin)
        {
            throw new NotImplementedException();
        }

        public override void SetPin(BluetoothAddress device, string pin)
        {
            throw new NotImplementedException();
        }

    }
}
